BASHPATH -xv

# the script creates sky subtracted images, using a model fit
# to a list of background values at empty ra dec positions

# $1: main directory
# $2: science directory
# $3: extension

SCRIPTDIR=`pwd`

INSTDIR=instruments_professional
if [ ! -f ${INSTDIR}/${INSTRUMENT}.ini ]; then
    INSTDIR=instruments_commercial
fi
if [ ! -f ${INSTDIR}/${INSTRUMENT}.ini ]; then
    INSTDIR=~/.theli/instruments_user/
fi
. ${INSTDIR}/${INSTRUMENT:?}.ini

. bash.include

# Check if file with blank sky positions exists
if [ ! -f /$1/$2/skypos.reg ]; then
    echo ERROR: Could not find /$1/$2/skypos.reg with blank sky positions
    exit
else
    # strip the ds9 region file, if necessary
    # this doesn't do anything if the file is already in correct format
    ${P_GAWK} '($0~/circle/)' /$1/$2/skypos.reg | sed 's/circle//g' | sed 's/[(),]/ /g' | awk '{if ($3~/"/) {print $1, $2, $3/60} else {print $1, $2, $3}}' > ${TEMPDIR}/skypos.dat_$$
    # Transform file to decimal format (if necessary).
    # Expand the center coordinates to min max square range
    # using the given box size (in arcmin; 1.0 arcmin if left empty)
    rm -f ${TEMPDIR}/my_skypos.dat_$$
    cd $1/$2
    exec < ${TEMPDIR}/skypos.dat_$$
    while read ra dec size
    do
	if [[ $ra =~ .*:.* ]]; then
	    ra=`hmstodecimal $ra`
	fi
	if [[ $dec =~ .*:.* ]]; then
	    dec=`dmstodecimal $dec`
	fi
	if [ "${size}"_A = "_A" ]; then
	    size=1
	fi
	ra1=`echo $ra $dec $size | ${P_GAWK} '{printf "%3.6f\n", $1-$3/60./2./cos($2*3.14159/180.)}'`
	ra2=`echo $ra $dec $size | ${P_GAWK} '{printf "%3.6f\n", $1+$3/60./2./cos($2*3.14159/180.)}'`
	dec1=`echo $dec $size    | ${P_GAWK} '{printf "%3.6f\n", $1-$2/60./2.}'`
	dec2=`echo $dec $size    | ${P_GAWK} '{printf "%3.6f\n", $1+$2/60./2.}'`
	# We don't need to care about poles at DEC=+/-90 degrees and the RA 0/360 border.
	# These are handled correctly in sky2xy already.
	echo ${ra1} ${dec1} ${ra2} ${dec2} ${ra} ${dec} >> ${TEMPDIR}/my_skypos.dat_$$
    done
fi

# Identify which are the images that contain the relevant regions
# We need to work on an exposure-by-exposure level (containing multiple chips)
ls -1 /$1/$2/*_1$3.fits > ${TEMPDIR}/skyfit_images_chip1_$$

exposures=`cat ${TEMPDIR}/skyfit_images_chip1_$$`
for file in $exposures
do
    base=`basename ${file} _1$3.fits`

    # Get a list of all associated exposures
    ls /$1/$2/${base}_*$3.fits > ${TEMPDIR}/skyfit_images_allchips_$$

    # Get the images that contain the largest relevant sky section
    # for each of the sky positions
    rm -f ${TEMPDIR}/${base}_skyfit.dat
    {
	exec < ${TEMPDIR}/my_skypos.dat_$$
	while read ra1 dec1 ra2 dec2 ra dec
	do
	    # don't do a line break here, it seems that goes wrong in the shell??
	    # Perhaps because of the nested execs / whiles
	    ${P_SKY2XY_HELPER} -i ${TEMPDIR}/skyfit_images_allchips_$$ -c ${ra1} ${dec1} ${ra2} ${dec2} > ${TEMPDIR}/skyfit_matchimages_$$
	    # Catch the case of region not found
	    bad=`grep "could not be found" ${TEMPDIR}/skyfit_matchimages_$$ | wc | ${P_GAWK} '{print $1}'`
	    if [ ${bad} == 0 ]; then
		image=`${P_GAWK} '{print $1}' ${TEMPDIR}/skyfit_matchimages_$$`
		section=`${P_GAWK} '{print $2, $3, $4, $5}' ${TEMPDIR}/skyfit_matchimages_$$`
		mode=`${P_IMSTATS} -d 6 ${image} -t -1e8 1e8 -s ${section} | ${P_GAWK} '($1 != "#") {print $2}'`
		echo "Region $ra $dec found in ${image}"
		echo ${ra} ${dec} ${mode} >> ${TEMPDIR}/${base}_skyfit.dat
	    fi
	done
    }
    
    # Now check how many data points we got identified
    # (incrementing inside nested loops in bash is nasty)
    ngood=`wc ${TEMPDIR}/${base}_skyfit.dat | ${P_GAWK} '{print $1}'`

    # This stuff only makes sense if we have a least three data points for a plane fit.
    # exit if less
    if [ ${ngood} -le 2 ]; then
	echo `${P_ERRTEST} 0`": Less than 3 sky positions were located in the images."
	exit
    fi
    
    # Exit if model is underconstrained
    nparam=`echo ${V_SKYFITORDER} | ${P_GAWK} '{print $1*2+1}'`
    if [ ${ngood} -lt ${nparam} ]; then
	echo `${P_ERRTEST} 0`": Insufficient number of sky positions for a model of degree ${V_SKYFITORDER}."
	exit
    fi
    
    # Now subtract the sky background models.
    # Not yet parallelized
    ${P_SKYFIT} \
	-i ${TEMPDIR}/${base}_skyfit.dat \
	-l ${TEMPDIR}/skyfit_images_allchips_$$ \
	-o ${V_SKYFITORDER}
    
done
