#include <fstream>
#include <iostream>
#include <sstream>
#include <cstdlib>
#include <vector>
#include <string>
#include <cstring>
#include <cmath>
#include <map>
#include <cctype>
#include <algorithm>

using namespace std;

void usage(int i)
{
  if (i == 0) {
    cerr << endl;
    cerr << "Date: 2013-04-30, Author: Mischa Schirmer\n" << endl;
    cerr << "Usage: background_dynamic " << endl;
    cerr << "       -i reference MJD" << endl;
    cerr << "       -f file with sky (or science) images (containing two columns with name and the MJD)" << endl;
    cerr << "       -w window size (from how many exposures do we create the background model?)" << endl;
    cerr << "      [-g maximum gap size in hours (default: 1.0)]\n" << endl;
    cerr << "Purpose:" << endl;
    cerr << "      Option -f is a file containing a list of (masked) images from which a background" << endl;
    cerr << "      model can be calculated. In addition to the file names the MJDs are tabulated as well." << endl;
    cerr << "      Option -i lists the reference MJD of an image that is to be corrected." << endl;
    cerr << "      The program will then select the n closest images (option -w) in time from the list" << endl;
    cerr << "      of masked images, excluding the corrected image should that be contained therein." << endl;
    cerr << "      Optionally, a gap size can be specified. If there is a gap larger than this value" << endl;
    cerr << "      in the series of masked images, then a warning is printed if the to be corrected exposure" << endl;
    cerr << "      is too far away, or an error is printed if the window covers the gap.\n" << endl;
  }

  exit (0);
}


//********************************************************  
int main(int argc, char *argv[])
{
  long win = 4;
  double ref_mjd = -1.0, gap = 1.0 / 24.;
  long i, j, n_sky;
  string sky_filename = "";
  string dummy;

  typedef struct {
    vector<string> name;
    vector<double> mjd;
    vector<double> index;
    vector<double> diff;
    vector<double> absdiff;
    vector<int> blockid;
  } data;

  struct data2 {
    string name;
    double mjd;
    int blockid;
  };
  
  map <double, data2> map_sky;
  
  // print usage if no arguments were given

  if (argc==1) usage(0);

  //  Read command line arguments
  
  for (i=1; i<argc; i++) {
    if (argv[i][0] == '-') {
      switch((int)argv[i][1]) {
      case 'i': ref_mjd = atof(argv[++i]);
	break;
      case 'f': sky_filename = argv[++i];
	break;
      case 'w': win = atol(argv[++i]);
	break;
      case 'g': gap = atof(argv[++i]) / 24.;
	break;
      }
    }
  }

  if (ref_mjd < 0.) {
    cerr <<  "ERROR: Reference MJD was not defined." << endl;
    exit (1);
  }

  if (win <= 2) {
    fprintf(stdout, "ERROR: %ld images used for dynamic background modelling. Must be at least 3!\n", win);
    exit (1);
  }

  // open the files
  const char *sky_file = sky_filename.c_str();
  ifstream input(sky_file);
  if (!input.is_open()) {
    cerr <<  "ERROR: Could not open " << sky_filename << "!\n";
    exit (1);
  }

  int blockcount = 0;
  double mjd_tmp, mjd_old = 0., diff;
  data skydata;
  data2 mydata;

  vector<double> mjd;
  
  j = 0;
  while (input >> dummy >> mjd_tmp) {
  //    istringstream iss(line);
    // Read filename and mjd
    // iss >> dummy >> mjd_tmp;
    // Leave if identical MJDs are found (the input list is ordered with respect to MJD)
    if (mjd_tmp == mjd_old) {
      cerr <<  "ERROR: Identical MJD entries found!" << endl;
      input.close();
      exit (1);
    }
    // Populate the structs and maps
    mydata.name = dummy;
    mydata.mjd = mjd_tmp;
    if (j == 0) mydata.blockid = 0;
    else {
      diff = mjd_tmp - mjd_old;
      if (diff > gap) blockcount++;
      mydata.blockid = blockcount;
    }
    // We drop the science exposure that is to be corrected
    if (ref_mjd == mjd_tmp) continue;

    map_sky[fabs(mjd_tmp-ref_mjd)] = mydata;
    mjd_old = mjd_tmp;
    mjd.push_back(mjd_tmp);
    j++;
  }

  n_sky = j;
  
  input.close();

  // If all exposures are taken after the reference MJD, 
  // with at least "gap" between MJD and the first exposure
  if (ref_mjd < mjd[0] && fabs(ref_mjd - mjd[0]) > gap) {
    cout << "NO_MATCHING_SKY_FOUND" << endl;
    exit (1);
  }

  // If all exposures are taken before the reference MJD, 
  // with at least "gap" between MJD and the last exposure
  if (ref_mjd > mjd[n_sky-1] && fabs(ref_mjd - mjd[n_sky-1]) > gap) {
    cout << "NO_MATCHING_SKY_FOUND" << endl;
    exit (1);
  }

  int count = 0;
  map <double, data2>::iterator mapit;
  for (mapit = map_sky.begin(); mapit != map_sky.end(); mapit++) {
    if (count++ < win) {
      cout << (*mapit).second.name << endl;
    }
  }

  // ADD a check to see whether sufficiently many background exposures were included.
  
  exit (0);

  /*

  int test_gap = 0;
  double found_gap = 0;
  int index_gap1 = 0, index_gap2 = 0;
  // test if there is a gap larger than the max gap size in the window
  for (j=unabs_start_index; j<unabs_start_index+win-1; j++) {
    //    printf("comp: %ld %s %s\n", j, skydata[s[j]].name, skydata[s[j+1]].name);
    if (fabs(skydata.mjd[s[j]] - skydata.mjd[s[j+1]]) > gap) {
      test_gap  = 1;
      found_gap = fabs(skydata.mjd[s[j]] - skydata.mjd[s[j+1]]);
      index_gap1 = s[j];
      index_gap2 = s[j+1];
    }
  }

  if (test_gap == 1) {
    cerr <<  "WARNING: The background window covers a gap of " << found_gap*24. << "hours between these exposures:" << endl;
    cerr <<  "         " << skydata.name[index_gap1] << endl;
    cerr <<  "         " << skydata.name[index_gap2] << endl;
    cerr <<  "         This is more than the maximum allowed gap size of " << gap*24. << "hours." << endl;
    //    exit (1);
  }

  // OK, no gap larger than the gap size!

  // Issue a warning if the window is larger than the gap size
  // This is not a problem as such, but it may mean that the number of images
  // is too large to sample the sky variations quickly enough
  //  for (j=unabs_start_index; j<unabs_start_index+win-1; j++) {
  double win_extent = fabs(skydata.mjd[unabs_start_index+s[win-1]] - skydata.mjd[unabs_start_index+s[0]]);
  if (win_extent > gap) {
    cerr <<  "WARNING: The background window length (" << win_extent*24 << ") hours" << endl;
    cerr <<  "         is larger than the maximum allowed gap size (" << gap*24 << ") hours" << endl;
    cerr <<  "         Sky variations might be sampled too \"slowly\"." << endl;
  }

  // print images, beginning with the closest image
  /*
  for (j=0; j<win; j++) {
    printf("%s\n", skydata[abs_s[j]].name);
  }

  // print images in window, sorted with respect to MJD
  for (j=0; j<win; j++) {
    cout << skydata.name[unabs_start_index+s[j]] << endl;
  }

  exit (0);
  */
  
}
